import os
import fnmatch
import shutil


def print_parse_exception(exc, filename=None):
    msg = "Parse Error "
    if filename:
        msg += "while compiling {0}".format(filename)
    msg += ": " + exc.msg + "\n"
    msg += exc.line + "\n"
    msg += " "*(exc.column-1) + "^"
    print(msg)


def walk_folder(root='.'):
    for subdir, dirs, files in os.walk(root):
        reldir = subdir[len(root):] if subdir.startswith(root) else subdir
        reldir = reldir.lstrip('/')
        for filename in files:
            yield os.path.join(reldir, filename)


def open_file(path, mode='rb', create_dir=False, create_mode=0o755):
    # Opens the given path. If create_dir is set, will
    # create all intermediate folders necessary to open
    try:
        newfile = open(path, mode)
    except IOError:
        # end here if not create_dir
        if not create_dir:
            raise
        newfile = None

    if not newfile:
        # may raise OSError
        filedir = os.path.split(path)[0]
        os.makedirs(filedir, create_mode)
        newfile = open(path, mode)

    return newfile


def copy_file(src, dst, create_dir=True, create_mode=0o755):
    try:
        shutil.copy2(src, dst)
    except IOError:
        if not create_dir:
            raise
        # may raise OSError
        filedir = os.path.split(dst)[0]
        os.makedirs(filedir, create_mode)
        shutil.copy2(src, dst)


def matches_pattern(pattern, filepath):

    def _is_match(pattern_list, token_list):
        if not pattern_list or not token_list:
            return False
        i, j = 0, 0
        while True:
            if pattern_list[j] == '**':
                if j+1 == len(pattern_list): return True
                if _is_match(pattern_list[j+1:], token_list[i:]):
                    return True
                else:
                    i+=1 
            elif fnmatch.fnmatch(token_list[i], pattern_list[j]):
                i+=1
                j+=1
            else:
                return False
            if i==len(token_list) and j==len(pattern_list):
                return True
            if i==len(token_list) or j==len(pattern_list):
                return False

    return _is_match(pattern.strip('/').split('/'), 
                     filepath.strip('/').split('/'))




from pyparsing import *
import inspect


# -----------------------------------------------------------------------------
# Exceptions
# -----------------------------------------------------------------------------

class TagErrorArguments(Exception):
    def __init__(self, tagname, nargs, args):
        params = (tagname, nargs, " ".join(args))
        errstr = "malformed tag '{0}' should have {1} argument(s), got '{2}'"
        self.msg = errstr.format(*params)

    def __str__(self):
        return self.msg


class TagErrorBody(Exception):
    def __init__(self, tagname, req_body, has_body):
        req = '' if req_body else "n't"
        has = 'does' if has_body else "doesn't"
        params = (tagname, req, has)
        errstr = "malformed tag '{0}' should{1} have a body, but {2}"
        self.msg = errstr.format(*params)

    def __str__(self):
        return self.msg


class TagErrorException(ParseBaseException):
    def __init__(self, parsestr, loc, exc, dev=False):
        if dev:
            import traceback
            msg = traceback.print_exc()
        else:
            msg = str(exc)
        super(TagErrorException, self).__init__(parsestr, loc=loc, msg=msg)


# -----------------------------------------------------------------------------
# Functions
# -----------------------------------------------------------------------------

def debug_action(name=''):
    def _wrapped(parsestr, loc, tokens):
        print(name, ": ", parsestr[0:loc], '*', parsestr[loc:], "-->", tokens)
    return _wrapped


# -----------------------------------------------------------------------------
# Classes
# -----------------------------------------------------------------------------

class TemplateLanguage(object):
    ''' A generic tag-based language supporting nested tags. 

    Each tag has a name, optional arguments, and an optional closing tag.

    Examples: 

    {% tagname arg1 arg2 %}

    {% tagname args %}
        tag body
    {% endtagname %}

    The action performed by each tag must be supplied in a tag definition,
    either passed to the __init__ method, or by using the add_tag 
    decorators.
    '''

    # decorators --------------------------------------------------------------

    def add_tag_with_name(self, name):
        ''' Adds a tag to the language. 

        The function for the decorator should 
        accept a list of positional args, the context keyword arg, and an
        optional body keyword arg.

        Example:

        @language.add_tag_with_name('mytag')
        def tagfn(arg1, arg2, body=u'', context={}):
            return "tag args: {0}, body: {1}".format([arg1, arg2], body])

        The positional args define the required arguments for the tag. 
        If the body keyword arg is present the tag must have a body and
        closing tag.

        Tag argument checking won't happen if the development flag is set.
        '''
        def _decorator(fn):
            posargs, varargs, varkwargs, defaults = inspect.getargspec(fn)
            req_body = "body" in posargs
            nargs = len(posargs)
            if "context" in posargs:
                nargs -= 1
            if req_body:
                nargs -= 1

            def _wrapper(*args, **kwargs):
                if not self._development:
                    if (varargs and len(args) < nargs) or len(args) != nargs:
                        raise TagErrorArguments(name, nargs, args)
                    has_body = 'body' in kwargs
                    if has_body != req_body:
                        raise TagErrorBody(name, req_body, has_body)
                return fn(*args, **kwargs)

            self._tags[name] = _wrapper

            return _wrapper
        return _decorator


    def add_tag(self, fn):
        ''' Shortcut for add_tag_with_name.

        Uses the function's name as the tag name.

        Example:

        @language.add_tag
        def mytag(body=u'', context={}):
            return "tag body: "+body
        '''
        return self.add_tag_with_name(fn.__name__)(fn)


    # language specification --------------------------------------------------

    def _mkopentag(self, name):
        tagname = CaselessKeyword(name)
        quote = quotedString.setParseAction(removeQuotes)
        arg = Optional(White()).suppress() + CharsNotIn(" \t\r\n")
        args = Group(ZeroOrMore(quote | arg))
        rawargs = SkipTo(self._tagclose)
        rawargs.setParseAction(lambda toks: args.parseString(toks[0]))
        return self._tagopen + tagname + rawargs + self._tagclose


    def _mkclosetag(self, name):
        tagname = CaselessKeyword("end"+name)
        return self._tagopen + tagname.suppress() + self._tagclose


    def _mktag(self, name, body):
        opentag = self._mkopentag(name)
        closetag = self._mkclosetag(name)
        tag = opentag + body + closetag | opentag
        return tag


    def _mkparser(self, tags):
        onechar = CharsNotIn('', exact=1)
        freetext = Combine(OneOrMore(~self._tagopen + ~self._tagclose + onechar))
        anytag = Forward()
        body = originalTextFor(ZeroOrMore(anytag | freetext))
        anytag << MatchFirst([self._mktag(key, body) for key in list(tags.keys())])
        return anytag


    def _mkparsefn(self, context):
        def _parsefn(parsestr, loc, tokens):
            name, parseresult = tokens[:2]
            args = parseresult.asList()
            try:
                fn = self._tags[name]
            except KeyError:
                # This should never be reached since the parser 
                # won't match a tag that's not in the list
                raise
            kwargs = {'context': context}
            if len(tokens) > 2:
                kwargs.update({'body': tokens[2]})
            try:
                processed = fn(*args, **kwargs)
            except ParseBaseException:
                raise
            except Exception as e:
                raise TagErrorException(parsestr, loc, e, self._development)
            return self.parse(processed, **context)
        return _parsefn


    # public methods ----------------------------------------------------------

    def __init__(self, tags=None, openseq='{%', closeseq='%}', development=False):
        ''' Creates a new template language instance.

        If the tag keyword argument isn't provided, tags should be created
        using the add_tag decorators. Otherwise, tags should be a dictionairy
        of name/function pairs.

        If the development flag is set, tag argument checking is disabled and
        errors will include a stack trace.
        '''
        self._tags = {}
        self._development = development
        self._openseq = openseq
        self._tagopen = Literal(openseq).suppress()
        self._tagclose = Literal(closeseq).suppress()

        if tags:
            for name, fn in tags.items():
                self.add_tag_with_name(name)(fn)
            self._parser = self._mkparser(self._tags)
        else:
            self._parser = None


    def parse(self, string, **context):
        ''' Parses a template string. 

        For each tag in the input string, calls the tag functions and replaces
        the tag with the function results. Keyword arguments provided to parse 
        will be added to the context passed to the tag functions.
        '''
        if self._openseq in string:
            if not self._parser:
                self._parser = self._mkparser(self._tags)
            parsefn = self._mkparsefn(context.copy())
            parser = self._parser.copy()
            parser.setParseAction(parsefn)
            return parser.transformString(string)
        else:
            return string




import os
from distutils.core import setup

root = os.path.dirname(os.path.realpath(__file__))

setup(
    name='brace-tags',
    version='1.0.10',
    author='Cole Krumbholz, Lauri Hynynen',
    author_email='team@brace.io',
    description='The simplest static site generator',
    packages=['tags'],
    install_requires=open(root+"/requirements.txt").read().splitlines(),
    long_description=open(root+"/README.md").read(),
    license='LICENSE',
    scripts=['scripts/tags'],
)


import os
import sys

from .templatelang import TemplateLanguage

lang = TemplateLanguage(openseq='{%', closeseq='%}')

@lang.add_tag
def include(path, context={}):
    '''
    Renders the content of a file. File paths should be relative
    to the site's root folder. Ex: {% include nav.html %}
    '''
    fullpath = os.path.join(context.get('rootdir'), path)
    if sys.version > '3':
        stuff = str(open(fullpath).read(), 'utf-8')
    else:
        stuff = unicode(open(fullpath).read(), 'utf-8')
    return stuff


@lang.add_tag_with_name('is')
def _is(path, body='', context={}):
    '''
    Renders the tag body if the path matches the current file. File paths 
    should be relative to the site's root folder. 
    Ex: {% is index.html %}Home!{% endis %}
    '''
    return body if path == context.get('filename') else ''


# ---- Add your custom tags here! -----

# Example:

# @lang.add_tag
# def print3x(style, body=u'', context={}):
#     ''' A tag that appends 3 copies of its body '''
#     result = body + body + body
#     if style == "bold":
#         result = u'<b>' + result + u'</b>'
#     return result

# Notes:

# - positional arguments define the tag's required arguments.
# - If you specify a `body` keyword argument, then the tag will require a body.
# - All tag functions must accept a `context` keyword argument. 

# You can also define tags that accept a variable argument list like so:

# @lang.add_tag
# def whatever(*args, **kwargs):
#     return str(len(args))


def render(content, filename='', rootdir='.'):
    ''' 
    Renders a content string containing template code into an output string. 
    Uses the tags specified above. Filename and rootdir are added to the 
    context passed to the tag functions.
    '''
    return lang.parse(content, filename=filename, rootdir=rootdir)
